package de.mdelab.mlsdm.interpreter.debug;

import java.io.OutputStream;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;

import de.mdelab.mlexpressions.MLExpression;
import de.mdelab.mlexpressions.mlcallactions.CallActionsExpressionLanguageConstants;
import de.mdelab.mlsdm.Activity;
import de.mdelab.mlsdm.ActivityEdge;
import de.mdelab.mlsdm.ActivityNode;
import de.mdelab.mlsdm.ActivityParameter;
import de.mdelab.mlsdm.ActivityParameterDirectionEnum;
import de.mdelab.mlsdm.interpreter.MLSDMInterpreter;
import de.mdelab.mlsdm.interpreter.callActions.CallActionInterpreter;
import de.mdelab.mlsdm.interpreter.coverage.CoverageReport;
import de.mdelab.mlsdm.interpreter.coverage.MLSDMCoverageNotificationReceiver;
import de.mdelab.mlsdm.interpreter.executionTrace.MLSDMExecutionTraceNotificationReceiver;
import de.mdelab.mlsdm.interpreter.expressions.MLSDMExpressionInterpreterManager;
import de.mdelab.mlsdm.interpreter.facade.MLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.facade.OptimizedMLSDMMetamodelFacadeFactory;
import de.mdelab.mlsdm.interpreter.notifications.MLSDMOutputStreamNotificationReceiver;
import de.mdelab.mlstorypatterns.AbstractStoryPatternLink;
import de.mdelab.mlstorypatterns.AbstractStoryPatternObject;
import de.mdelab.mlstorypatterns.StoryPattern;
import de.mdelab.sdm.interpreter.core.SDMException;
import de.mdelab.sdm.interpreter.core.SDMInterpreter;
import de.mdelab.sdm.interpreter.core.debug.SDCoverageNotificationReceiver;
import de.mdelab.sdm.interpreter.core.debug.SDDebugger;
import de.mdelab.sdm.interpreter.core.debug.SDExecution;
import de.mdelab.sdm.interpreter.core.debug.SDInterpreterLauncher;
import de.mdelab.sdm.interpreter.core.executionTrace.ExecutionTrace;
import de.mdelab.sdm.interpreter.core.executionTrace.ExecutionTraceNotificationReceiver;
import de.mdelab.sdm.interpreter.core.expressions.ExpressionInterpreterManager;
import de.mdelab.sdm.interpreter.core.notifications.OutputStreamNotificationReceiver;
import de.mdelab.sdm.interpreter.core.variables.Variable;
import de.mdelab.sdm.interpreter.ocl.OCLExpressionInterpreter;
import de.mdelab.sdm.interpreter.ocl.OCLExpressionInterpreterConstants;

public class MLSDMInterpreterLauncher extends SDInterpreterLauncher<CoverageReport, Activity, EClassifier> {

	private boolean useOptimizations;
	
	public MLSDMInterpreterLauncher() {
		super();
		
		useOptimizations = false;
	}
	
	public MLSDMInterpreterLauncher( final ClassLoader p_classLoader ) {
		super( p_classLoader );
	}
	
	@Override
	public MLSDMExecution launch( 	final Activity p_activity,
									final List<Variable<EClassifier>> p_variableBindings,
									final Integer p_debuggerPort,
									final Integer p_resourceSetSynchronizerPort,
									final boolean pb_debugOutput,
									final boolean pb_execTrace,
									final boolean pb_coverageTrace,
									final IProgressMonitor p_monitor,
									final Logger p_logger,
									final OutputStream p_outputStream )
	throws SDMException {
		return this.launch( p_activity, p_variableBindings, p_debuggerPort, p_resourceSetSynchronizerPort, pb_debugOutput, pb_execTrace, pb_coverageTrace, false, p_monitor, p_logger, p_outputStream );
	}
	
	public MLSDMExecution launch( 	final Activity p_activity,
									final List<Variable<EClassifier>> p_variableBindings,
									final Integer p_debuggerPort,
									final Integer p_resourceSetSynchronizerPort,
									final boolean pb_debugOutput,
									final boolean pb_execTrace,
									final boolean pb_coverageTrace,
									final boolean pb_useOptimizations,
									final IProgressMonitor p_monitor,
									final Logger p_logger,
									final OutputStream p_outputStream )
	throws SDMException {
		useOptimizations = pb_useOptimizations;
		
		return (MLSDMExecution) super.launch( p_activity, p_variableBindings, p_debuggerPort, p_resourceSetSynchronizerPort, pb_debugOutput, pb_execTrace, pb_coverageTrace, p_monitor, p_logger, p_outputStream );
	}

	private MLSDMInterpreter createInterpreter(	final ExpressionInterpreterManager<Activity, ActivityNode, ActivityEdge, StoryPattern, AbstractStoryPatternObject, AbstractStoryPatternLink, EClassifier, EStructuralFeature, MLExpression> p_expressionManager,
			 									final OptimizedMLSDMMetamodelFacadeFactory p_facadeFactory ) {
		return new MLSDMInterpreter( p_expressionManager, p_facadeFactory );
	}
	
	@Override
	protected SDMInterpreter<Activity, ?, ?, ?, ?, ?, EClassifier, ?, ?> createInterpreter( final List<Variable<EClassifier>> p_variableBindings )
	throws SDMException {
		if ( useOptimizations ) {
			final Set<EPackage> modelPackages = new HashSet<EPackage>();
			final Set<EObject> models = new HashSet<EObject>();
			
			for ( final Variable<EClassifier> var : p_variableBindings ) {
				modelPackages.add( var.getClassifier().getEPackage() );

				final Object varVal = var.getValue();
				
				if ( varVal instanceof EObject ) {
					models.add( EcoreUtil.getRootContainer( (EObject) varVal ) );
				}
			}
			
			final OptimizedMLSDMMetamodelFacadeFactory facadeFactory = new OptimizedMLSDMMetamodelFacadeFactory( modelPackages );
			facadeFactory.getInstanceFacade().initialize( models );
			
			final MLSDMInterpreter interpreter = createInterpreter( facadeFactory );
			final CallActionInterpreter actInterpreter = new CallActionInterpreter() {
				
				@Override
				protected MLSDMInterpreter createMLSDMInterpreter() {
					return createInterpreter( interpreter.getExpressionInterpreterManager(), facadeFactory );
				}
			};

			interpreter.getExpressionInterpreterManager().registerExpressionInterpreter( 	actInterpreter,
																							CallActionsExpressionLanguageConstants.CALL_ACTION_EXPRESSION_LANGUAGE_NAME, 
																							CallActionsExpressionLanguageConstants.CALL_ACTION_EXPRESSION_LANGUAGE_VERSION );
			return interpreter;
		}

		return createInterpreter( MLSDMMetamodelFacadeFactory.INSTANCE );
	}
	
	protected MLSDMInterpreter createInterpreter( final MLSDMMetamodelFacadeFactory p_facadeFactory )
	throws SDMException {
		final MLSDMInterpreter interpreter = new MLSDMInterpreter( new MLSDMExpressionInterpreterManager( classLoader ), p_facadeFactory );
		interpreter.getExpressionInterpreterManager().registerExpressionInterpreter(	new CallActionInterpreter(),
																						CallActionsExpressionLanguageConstants.CALL_ACTION_EXPRESSION_LANGUAGE_NAME, 
																						CallActionsExpressionLanguageConstants.CALL_ACTION_EXPRESSION_LANGUAGE_VERSION );
		interpreter.getExpressionInterpreterManager().registerExpressionInterpreter(	new OCLExpressionInterpreter<MLExpression>(),
																						OCLExpressionInterpreterConstants.OCL_LANGUAGE_NAME,
																						OCLExpressionInterpreterConstants.OCL_LANGUAGE_VERSION );
		
		return interpreter;
	}

	@Override
	protected SDDebugger<?, Activity, ?, ?, ?> createDebugger( 	final SDMInterpreter<Activity, ?, ?, ?, ?, ?, EClassifier, ?, ?> p_sdi,
																final Integer p_debuggerPort, 
																final Integer p_resourceSyncPort) {
		return new MLSDMDebugger( (MLSDMInterpreter) p_sdi, p_debuggerPort, p_resourceSyncPort );
	}

	@Override
	protected ExecutionTraceNotificationReceiver<Activity, ?, ?, ?, ?, ?, EClassifier, ?, ?> createExecutionTraceNotificationReceiver() {
		return new MLSDMExecutionTraceNotificationReceiver();
	}

	@Override
	protected OutputStreamNotificationReceiver<Activity, ?, ?, ?, ?, ?, EClassifier, ?, ?> createOutputStreamNotificationReceiver( final OutputStream p_outputStream ) {
		return  new MLSDMOutputStreamNotificationReceiver( p_outputStream );
	}

	@Override
	protected SDCoverageNotificationReceiver<CoverageReport, EClassifier> createCoverageNotificationReceiver() {
		return new MLSDMCoverageNotificationReceiver();
	}

	@Override
	protected String getName( final Activity p_activity ) {
		assert p_activity != null;
		
		return p_activity.getName();
	}

	@Override
	protected void logOutputParameterValues(	final Activity p_activity,
												final Map<String, Variable<EClassifier>> p_returnVariables,
												final Logger p_logger ) {
		for ( final ActivityParameter param : p_activity.getParameters() ) {
			final ActivityParameterDirectionEnum direction = param.getDirection();
			
			if ( direction == ActivityParameterDirectionEnum.OUT || direction == ActivityParameterDirectionEnum.INOUT ) {
				p_logger.info( "Output parameter '"	+ param.getName() + "' : " + p_returnVariables.get( param.getName() ).getValue() );
			}
		}
	}

	@Override
	public MLSDMDebugger getDebugger() {
		return (MLSDMDebugger) super.getDebugger();
	}

	@Override
	protected SDExecution<CoverageReport, EClassifier> createExecutionResult( 	final ExecutionTrace p_execTrace,
																				final CoverageReport p_covReport,
																				final Map<String, Variable<EClassifier>> p_returnVariables ) {
		return new MLSDMExecution( p_execTrace, p_covReport, p_returnVariables );
	}
}
